mod mutation;
mod query;
mod schema;
mod subscription;

use std::net::SocketAddr;

use async_graphql::{
    http::{playground_source, GraphQLPlaygroundConfig},
    Request, Response, Schema,
};
use async_graphql_axum::GraphQLSubscription;
use axum::{
    http::Method,
    response::{Html, IntoResponse},
    routing, Extension, Json, Router,
};
use mutation::MutationRoot;
use query::QueryRoot;
use schema::ScelSchema;
use subscription::SubscriptionRoot;
use tower_http::cors::CorsLayer;

async fn graphql_playground() -> impl IntoResponse {
    Html(playground_source(
        GraphQLPlaygroundConfig::new("/").subscription_endpoint("/ws"),
    ))
}
async fn graphql_handler(schema: Extension<ScelSchema>, req: Json<Request>) -> Json<Response> {
    schema.execute(req.0).await.into()
}

pub struct Server {
    app: Router,
    addr: SocketAddr,
}

impl Server {
    pub fn new() -> Server {
        let schema = Schema::build(QueryRoot, MutationRoot, SubscriptionRoot).finish();

        let cors = vec!["http://localhost:3000"
            .parse()
            .expect("Could not parse url")];

        let app = Router::new()
            .route("/", routing::get(graphql_playground).post(graphql_handler))
            .route("/ws", GraphQLSubscription::new(schema.clone()))
            .layer(Extension(schema))
            .layer(
                CorsLayer::new()
                    .allow_origin(cors)
                    .allow_headers([axum::http::header::CONTENT_TYPE])
                    .allow_methods([Method::GET, Method::POST, Method::OPTIONS]),
            );
        let addr = SocketAddr::from(([127, 0, 0, 1], 3000));

        return Server { app, addr };
    }

    pub async fn start(self) -> anyhow::Result<()> {
        tracing::info!("listening on {}", self.addr);

        match axum::Server::bind(&self.addr)
            .serve(self.app.into_make_service())
            .await
        {
            Ok(_) => Ok(()),
            Err(e) => Err(e.into()),
        }
    }
}
